/*
 * spi.c
 *
 * SPI (M25P64) I/f
 *
 * Copyright (C) 2006 Atmark Techno, Inc.  All Rights Reserved.
 *
 */

#include <xparameters.h>

#ifdef XPAR_SPI_FLASH_BASEADDR

#include "memmap.h"
#include "medium.h"
#include "spi.h"

/* Register */
#if 0 /* unused */
#define SPI_INPUT_GLOBAL_ENABLE_REG ((*(volatile unsigned long *) (SPI_REGISTER_BASEADDR + 0x1C)))
#define SPI_INTERRUPT_REG           ((*(volatile unsigned long *) (SPI_REGISTER_BASEADDR + 0x20)))
#define SPI_INTERRUPT_ENABLE_REG    ((*(volatile unsigned long *) (SPI_REGISTER_BASEADDR + 0x28)))
#define SPI_RESET_MODULE            ((*(volatile unsigned long *) (SPI_REGISTER_BASEADDR + 0x40)))
#endif
#define SPI_CONTROL_REG             ((*(volatile unsigned long *) (SPI_REGISTER_BASEADDR + 0x060)))
#define SPI_STATUS_REG              ((*(volatile unsigned long *) (SPI_REGISTER_BASEADDR + 0x064)))
#define SPI_DATA_TRANSMIT_REG       ((*(volatile unsigned long *) (SPI_REGISTER_BASEADDR + 0x068)))
#define SPI_DATA_RECEIVE_REG        ((*(volatile unsigned long *) (SPI_REGISTER_BASEADDR + 0x06C)))
#define SPI_SLAVE_SELECT_REG        ((*(volatile unsigned long *) (SPI_REGISTER_BASEADDR + 0x070)))
#if 0 /* unused */
#define SPI_TRANSMIT_FIFO_OCCUPANCY ((*(volatile unsigned long *) (SPI_REGISTER_BASEADDR + 0x074)))
#define SPI_RECEIVE_FIFO_OCCUPANCY  ((*(volatile unsigned long *) (SPI_REGISTER_BASEADDR + 0x078)))
#endif

/* SPI Instruction set */
#define SPI_CODE_WREN       0x06    /* Write Enable */
#define SPI_CODE_WRDI       0x04    /* Write Disable */
#define SPI_CODE_RDID       0x9F    /* Read Identification */
#define SPI_CODE_RDSR       0x05    /* Read Status Register */
#define SPI_CODE_WRSR       0x01    /* Write Status Register */
#define SPI_CODE_READ       0x03    /* Read Data Bytes */
#define SPI_CODE_FAST_READ  0x0B    /* Read Data Bytes at Higher Speed */
#define SPI_CODE_PP         0x02    /* Page Program */
#define SPI_CODE_SE         0xD8    /* Sector Erase */
#define SPI_CODE_BE         0xC7    /* Bulk Erase */
#define SPI_CODE_RES        0xAB    /* Read Electronic Signature */

/* SPI Control */
#define SPI_CONTROL_LOOP                        0x001   /* Local loopback mode */
#define SPI_CONTROL_SPE                         0x002   /* SPI System Enable select bit */
#define SPI_CONTROL_MSTR                        0x004   /* MSTR */
#define SPI_CONTROL_CPOL                        0x008   /* Clock Polarity select bit */
#define SPI_CONTROL_CPHA                        0x010   /* Clocl Phase select bit */
#define SPI_CONTROL_TX_FIFO_RESET               0x020   /* Transmit FIFO Reset */
#define SPI_CONTROL_RX_FIFO_RESET               0x040   /* Receive FIFO Reset */
#define SPI_CONTROL_MANUAL_SS_ASSERTION_ENABLE  0x080   /* Manual SS Assertion Enable */
#define SPI_CONTROL_MASTER_TRANSATION_INHIBIT   0x100   /* Master Transaction Inhibit */

/* SPI Status */
#define SPI_STATUS_RC_EMPTY 0x01    /* Receive Empty */
#define SPI_STATUS_RC_FULL  0x02    /* Receive Full */
#define SPI_STATUS_TX_EMPTY 0x04    /* Transmit Empty */
#define SPI_STATUS_TX_FULL  0x08    /* Transmit Full */
#define SPI_STATUS_MODF     0x10    /* Mode-Fault Error Flag */

/* Reboot */
#define SYSTEM_GPIO_BASE (SYSTEM_REGISTER_BASEADDR)
#define SYSTEM_GPIO(v)   (*(volatile unsigned long *)(SYSTEM_GPIO_BASE) = (v))
#define REBOOT           (1)

/* Functions */
static void SPI_Initialize(void);
static void SPI_WriteEnable(void);
static void SPI_WriteDisable(void);
static void SPI_ReadIdentification(void);
static void SPI_ReadStatusRegister(void);
#if 0 /* unused */
static void SPI_WriteStatusRegister(void);
#endif
static void SPI_ReadDataBytes(void);
static void SPI_FastReadDataBytes(unsigned long addr, void * buffer, int length);
static void SPI_PageProgram(unsigned long addr, void * data);
static void SPI_SectorErase(unsigned long addr);
static void SPI_BulkErase(void);
static int  SPI_ReadElectronicSignature(void);

/*
 *  SPI Copy from DRAM
 */
void spi_copy_from_dram(unsigned long start_addr, unsigned long last_addr)
{
        unsigned long offset_start, offset_last, offset;
        int i;

        /* Initialize SPI */
        SPI_Initialize();

        /* Erase SPI */
        offset_start = ((start_addr) & 0x7F0000);
        offset_last  = ((last_addr) & 0x7F0000);
        myprint("\r\nErasing SPI ...\r\n");

        for (offset = offset_start; offset <= offset_last; offset += 0x10000) {
                /* Erase SPI */
                SPI_WriteEnable();
                SPI_SectorErase( offset );
                SPI_ReadStatusRegister();
        }

        /* Copy SDRAM to SPI */
        offset_start = ((start_addr) & 0x7FFF00);
        offset_last  = ((last_addr + 0x100) & 0x7FFF00);
        myprint("Programming SPI ...\r\n");

        for (offset = offset_start; offset < offset_last; offset += 0x100) {
                /* Program SPI */
                SPI_WriteEnable();
                SPI_PageProgram(offset, (void *)(SDRAM_BASEADDR | (offset & 0x7FFFFF)) );
                SPI_ReadStatusRegister();
        }
        myprint("done.\r\n");
        myprint("Reboot.\r\n");
        for (i = 0; i < 2000000000; i++) {}
        SYSTEM_GPIO(REBOOT);
}

/*
 *  SPI Copy to DRAM
 */
void spi_copy_to_dram(unsigned long addr, void *buffer, int size)
{
        unsigned long offset;

        /* Initialize SPI */
        SPI_Initialize();

        /* Read SPI and Copy to DRAM */
        for (offset = 0x00000; offset < size; offset += 0x100) {
                SPI_FastReadDataBytes(addr + offset,
                                      (void *)(buffer + offset), 0x100);
        }
}

/******************************************************************************
 *  SPI (M25P64) I/f
 */

/*
 *  Initialize
 */
void SPI_Initialize(void)
{
        /* Fifo reset Enable */
        SPI_CONTROL_REG =
                SPI_CONTROL_MASTER_TRANSATION_INHIBIT |
                SPI_CONTROL_MANUAL_SS_ASSERTION_ENABLE |
                SPI_CONTROL_RX_FIFO_RESET |
                SPI_CONTROL_TX_FIFO_RESET;

        /* Fifo reset Disable */
        SPI_CONTROL_REG =
                SPI_CONTROL_MASTER_TRANSATION_INHIBIT |
                SPI_CONTROL_MANUAL_SS_ASSERTION_ENABLE;

        /* Slave select */
        SPI_SLAVE_SELECT_REG = 0xfe;
}

/*
 *  Write Enable(SPI_CODE_WREN)
 */
void SPI_WriteEnable(void)
{
        /* CS Enable */
        SPI_CONTROL_REG =
                SPI_CONTROL_MANUAL_SS_ASSERTION_ENABLE |
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;

        /* Write Enable Command */
        SPI_DATA_TRANSMIT_REG = SPI_CODE_WREN;

        /* OPB SPI Status FIFO TX end wait */
        while ((SPI_STATUS_REG & SPI_STATUS_TX_EMPTY) == 0);

        /* CS Desable */
        SPI_CONTROL_REG =
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;
}

/*
 *  Write Disable(SPI_CODE_WRDI)
 */
void SPI_WriteDisable(void)
{
        /* CS Enable */
        SPI_CONTROL_REG =
                SPI_CONTROL_MANUAL_SS_ASSERTION_ENABLE |
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;

        /* Write Disable Command */
        SPI_DATA_TRANSMIT_REG = SPI_CODE_WRDI;

        /* OPB SPI Status FIFO TX end wait */
        while ((SPI_STATUS_REG & SPI_STATUS_TX_EMPTY) == 0);

        /* CS Desable */
        SPI_CONTROL_REG =
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;
}

/*
 *  Read Identification(SPI_CODE_RDID)
 */
void SPI_ReadIdentification(void)
{
}

/*
 *  Read Status Register(SPI_CODE_RDSR)
 */
void SPI_ReadStatusRegister(void)
{
        int tmp;

        /* Fifo reset Enable CS Disable */
        SPI_CONTROL_REG =
                SPI_CONTROL_RX_FIFO_RESET |
                SPI_CONTROL_TX_FIFO_RESET |
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;

        /* CS Enable */
        SPI_CONTROL_REG =
                SPI_CONTROL_MANUAL_SS_ASSERTION_ENABLE |
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;

        /* Read Status Register Command */
        SPI_DATA_TRANSMIT_REG = SPI_CODE_RDSR;

        /* OPB SPI Status FIFO TX end wait */
        while ((SPI_STATUS_REG & SPI_STATUS_TX_EMPTY) == 0);

        /* Dummy read */
        tmp = SPI_DATA_RECEIVE_REG;

        while (1) {
                /* Dummy Write */
                SPI_DATA_TRANSMIT_REG = 0x00;
                /* OPB SPI Status */
                while (SPI_STATUS_REG & SPI_STATUS_RC_EMPTY);
                /* SPI Status */
                if ((SPI_DATA_RECEIVE_REG & 0x1) == 0)
                        break;
        }

        /* CS Desable */
        SPI_CONTROL_REG =
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;
}

#if 0 /* unused */
/*
 *  Write Status Register(SPI_CODE_WRSR)
 */
void SPI_WriteStatusRegister(void)
{
        /* Fifo reset Enable CS Disable */
        SPI_CONTROL_REG =
                SPI_CONTROL_RX_FIFO_RESET |
                SPI_CONTROL_TX_FIFO_RESET |
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;

        /* CS Enable */
        SPI_CONTROL_REG =
                SPI_CONTROL_MANUAL_SS_ASSERTION_ENABLE |
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;

        SPI_WriteEnable();

        /* Fifo reset Enable CS Disable */
        SPI_CONTROL_REG =
                SPI_CONTROL_RX_FIFO_RESET |
                SPI_CONTROL_TX_FIFO_RESET |
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;

        SPI_ReadStatusRegister();

        put_char_hex(SPI_DATA_RECEIVE_REG);
        put_char_hex(SPI_DATA_RECEIVE_REG);
        put_char_hex(SPI_DATA_RECEIVE_REG);

        /* CS Enable */
        SPI_CONTROL_REG =
                SPI_CONTROL_MANUAL_SS_ASSERTION_ENABLE |
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;

        /* Write Status Register Command */
        SPI_DATA_TRANSMIT_REG = SPI_CODE_WRSR;

        /* OPB SPI Status FIFO TX end wait */
        while ((SPI_STATUS_REG & SPI_STATUS_TX_EMPTY) == 0);

        /* Write Status Register Command */
        SPI_DATA_TRANSMIT_REG = 0x0;

        /* OPB SPI Status FIFO TX end wait */
        while ((SPI_STATUS_REG & SPI_STATUS_TX_EMPTY) == 0);

        /* CS Disable */
        SPI_CONTROL_REG =
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;
}
#endif

/*
 *  Read Data Bytes(SPI_CODE_READ)
 */
void SPI_ReadDataBytes(void)
{
}

/*
 *  Read Data Bytes at Higher Speed(SPI_CODE_FAST_READ)
 */
void SPI_FastReadDataBytes(unsigned long addr, void * buffer, int length)
{
        unsigned char *data = (unsigned char *)buffer ;
        int i;

        /* Fifo reset Enable CS Disable */
        SPI_CONTROL_REG =
                SPI_CONTROL_RX_FIFO_RESET |
                SPI_CONTROL_TX_FIFO_RESET |
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;

        /* CS Enable */
        SPI_CONTROL_REG =
                SPI_CONTROL_MANUAL_SS_ASSERTION_ENABLE |
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;

        /* Read Command */
        SPI_DATA_TRANSMIT_REG = SPI_CODE_FAST_READ;

        SPI_DATA_TRANSMIT_REG = (addr>>16) & 0xFF;    /* Addr 2 */
        SPI_DATA_TRANSMIT_REG = (addr>> 8) & 0xFF;    /* Addr 1 */
        SPI_DATA_TRANSMIT_REG = (addr>> 0) & 0xFF;    /* Addr 0 */

        /* OPB SPI Status FIFO TX end wait */
        while ((SPI_STATUS_REG & 0x4) == 0);

        /* RX Fifo reset Enable CS Enable */
        SPI_CONTROL_REG =
                SPI_CONTROL_MANUAL_SS_ASSERTION_ENABLE |
                SPI_CONTROL_RX_FIFO_RESET |
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;

        /* RX Fifo reset Disable CS Enable */
        SPI_CONTROL_REG =
                SPI_CONTROL_MANUAL_SS_ASSERTION_ENABLE |
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;

        SPI_DATA_TRANSMIT_REG = 0x00;     /* Dummy Write */

        /* OPB SPI Status FIFO TX end wait */
        while ((SPI_STATUS_REG & SPI_STATUS_TX_EMPTY) == 0);

        i = SPI_DATA_RECEIVE_REG;    /* Dummy read */
        for (i = 0; i < length; i++) {
                SPI_DATA_TRANSMIT_REG = 0x00;    /* Dummy Write */
                while (SPI_STATUS_REG & SPI_STATUS_RC_EMPTY);
                data[i] = SPI_DATA_RECEIVE_REG;    /* Dummy read */
        }

        /* CS Disable */
        SPI_CONTROL_REG =
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;
}

/*
 *  Page Program(SPI_CODE_PP)
 */
void SPI_PageProgram(unsigned long addr, void * data)
{
        unsigned char * buffer = (unsigned char *) data;
        int i;

        /* CS Enable */
        SPI_CONTROL_REG =
                SPI_CONTROL_MANUAL_SS_ASSERTION_ENABLE |
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;

        /* Write Enable Command */
        SPI_DATA_TRANSMIT_REG = SPI_CODE_WREN;

        /* OPB SPI Status FIFO TX end wait */
        while ((SPI_STATUS_REG & SPI_STATUS_TX_EMPTY) == 0);

        /* CS Disable */
        SPI_CONTROL_REG =
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;

        /* Fifo reset Enable CS Disable */
        SPI_CONTROL_REG =
                SPI_CONTROL_RX_FIFO_RESET |
                SPI_CONTROL_TX_FIFO_RESET |
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;

        /* Fifo reset Disable CS Enable */
        SPI_CONTROL_REG =
                SPI_CONTROL_MANUAL_SS_ASSERTION_ENABLE |
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;

        SPI_DATA_TRANSMIT_REG = SPI_CODE_PP;        /* Page Program Command */
        SPI_DATA_TRANSMIT_REG = (addr>>16) & 0xFF;  /* Addr 2 */
        SPI_DATA_TRANSMIT_REG = (addr>> 8) & 0xFF;  /* Addr 1 */
        SPI_DATA_TRANSMIT_REG = (addr>> 0) & 0xFF;  /* Addr 0 */

        for (i = 0; i < 256; i++) {
                SPI_DATA_TRANSMIT_REG = buffer[i];    /* write data */
                /* OPB SPI Status FIFO TX end wait */
                while ((SPI_STATUS_REG & SPI_STATUS_TX_FULL) != 0);
        }

        /* OPB SPI Status FIFO TX end wait */
        while ((SPI_STATUS_REG & SPI_STATUS_TX_EMPTY) == 0);

        /* CS Disable */
        SPI_CONTROL_REG =
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;
}

/*
 *  Sector Erase(SPI_CODE_SE)
 */
void SPI_SectorErase(unsigned long addr)
{
        /* Fifo reset Enable CS Disable */
        SPI_CONTROL_REG =
                SPI_CONTROL_RX_FIFO_RESET |
                SPI_CONTROL_TX_FIFO_RESET |
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;

        /* Fifo reset Disable CS Enable */
        SPI_CONTROL_REG =
                SPI_CONTROL_MANUAL_SS_ASSERTION_ENABLE |
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;

        SPI_DATA_TRANSMIT_REG = SPI_CODE_SE;          /* Sector Erase Comand */
        SPI_DATA_TRANSMIT_REG = (addr>>16) & 0xFF;    /* Addr 2 */
        SPI_DATA_TRANSMIT_REG = (addr>> 8) & 0xFF;    /* Addr 1 */
        SPI_DATA_TRANSMIT_REG = (addr>> 0) & 0xFF;    /* Addr 0 */

        /* OPB SPI Status FIFO TX end wait */
        while ((SPI_STATUS_REG & SPI_STATUS_TX_EMPTY) == 0);

        /* CS Disable */
        SPI_CONTROL_REG =
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;
}

/*
 *  Bulk Erase(SPI_CODE_BE)
 */
void SPI_BulkErase(void)
{
        /* Fifo reset Enable CS Disable */
        SPI_CONTROL_REG =
                SPI_CONTROL_RX_FIFO_RESET |
                SPI_CONTROL_TX_FIFO_RESET |
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;

        /* Fifo reset Disable CS Enable */
        SPI_CONTROL_REG =
                SPI_CONTROL_MANUAL_SS_ASSERTION_ENABLE |
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;

        /* Write Enable Command */
        SPI_DATA_TRANSMIT_REG = SPI_CODE_BE;

        /* OPB SPI Status FIFO TX end wait */
        while ((SPI_STATUS_REG & SPI_STATUS_TX_EMPTY) == 0);

        /* CS Disable */
        SPI_CONTROL_REG =
                SPI_CONTROL_MSTR |
                SPI_CONTROL_SPE;
}

/*
 *  Read Electronic Signature(SPI_CODE_RES)
 */
int SPI_ReadElectronicSignature(void)
{
        return -1;
}

#endif /* XPAR_SPI_FLASH_BASEADDR */
